#!/usr/bin/env python
"""
Setup a virtual machine from a config file.

Note: Most of the operations requires root access and this script uses ``sudo``
to get them.

"""
import os
import subprocess
import sys
import tempfile
import time


from bzrlib import (
    config,
    osutils,
    transport,
    urlutils,
    )

# Work around a bug in bzrlib.config forbidding some constructs in templates.
# Namely, spaces are invalid as an identifier and therefore should not match
# below.
config._option_ref_re = config.lazy_regex.lazy_compile('({[^ {}\n]+})')

class VmMatcher(config.NameMatcher):

    def match(self, section):
        if section.id is None:
            # The no name section contains default values
            return True
        return super(VmMatcher, self).match(section)

    def get_sections(self):
        matching_sections = super(VmMatcher, self).get_sections()
        return reversed(list(matching_sections))


class VmStore(config.LockableIniFileStore):
    """A config store for options specific to a directory."""

    def __init__(self, directory, file_name, possible_transports=None):
        t = transport.get_transport_from_path(
            directory, possible_transports=possible_transports)
        super(VmStore, self).__init__(t, file_name)
        self.id = 'vm'


class VmStack(config.Stack):
    """Per-directory options."""

    def __init__(self, name):
        """Make a new stack for a given vm.

        :param name: The name of a virtual machine."""
        dstore = VmStore(os.environ['HOME'], 'vms.conf')
        # FIXME: Only available in bzr-2.6b3 :-/ -- vila 2012-01-31
        # dstore = self.get_shared_store()
        super(VmStack, self).__init__(
            [VmMatcher(dstore, name).get_sections,],
            dstore, mutable_section_id=name)


def register(option):
    config.option_registry.register(option)


# The base directories where vms are stored
register(config.Option(
        'vm.images_dir', default='/var/lib/libvirt/images',
        help="Where vm disk images are stored."
        ))
register(config.Option(
        'vm.qemu_etc_dir',
        default='/etc/libvirt/qemu',
        help="Where libvirt (qemu) stores the vms config files."
        ))

# Isos and images download handling
register(config.Option(
        'vm.iso_url',
        default='http://cdimage.ubuntu.com/daily-live/current/' ,
        help="Where an iso can be downloaded from."
        ))
register(config.Option(
        'vm.iso_name',
        default='{vm.release}-desktop-{vm.cpu_model}.iso',
        help="The name of the iso."
        ))
register(config.Option(
        'vm.cloud_image_url',
        default='http://cloud-images.ubuntu.com/{vm.release}/current/',
        help="Where a cloud image can be downloaded from."
        ))
register(config.Option(
        'vm.cloud_image_name',
        default='{vm.release}-server-cloudimg-{vm.cpu_model}-disk1.img',
        help="The name of the cloud image."
        ))
register(config.Option(
        'vm.download_cache',
        default=None,
        help="Where downloads end up."
        ))

# The options required to create a vm
register(config.Option(
        'vm.name', default=None, invalid='error',
        help="The vm name, used as a prefix for related files."
        ))
register(config.Option(
        'vm.ram_size', default='1024',
        help="The ram size in megabytes."
        ))
register(config.Option(
        'vm.disk_size', default='8G',
        help="The ram size in megabytes."
        ))
register(config.Option(
        'vm.cpus', default='1',
        help="The number of cpus."
        ))
register(config.Option(
        'vm.cpu_model', default=None,
        help="The number of cpus."))
register(config.Option(
        'vm.password', default='ubuntu', invalid='error',
        help="The ubuntu user password."
        ))
register(config.Option(
        'vm.release', default=None, invalid='error',
        help="The ubuntu release name."
        ))
register(config.Option(
        'vm.meta_data', default='''\
instance-id: nocloud
local-hostname: {vm.name}
''',
        invalid='error',
        help="The meta data for cloud-init to put in the seed."
        ))
register(config.Option(
        'vm.user_data', default='''\
#cloud-config
locale: en_US.UTF-8

password: {vm.password}
chpasswd: { expire: False }
ssh_pwauth: True

ssh_authorized_keys:
 - {vm.ssh_pub_key}

#apt_proxy: http://10.189.74.2:8000
apt_sources: 
 - source: deb http://archive.ubuntu.com/ubuntu {vm.release} restricted multiverse
 - source: deb-src http://archive.ubuntu.com/ubuntu {vm.release} restricted multiverse

apt_update: true
apt_upgrade: true
byobu_by_default: system

manage_etc_hosts: true

packages:
 - bzr
 - ubuntu-desktop
 - openssh-server
 - avahi-daemon

''',
        invalid='error',
        help="The user data for cloud-init to put in the seed."
        ))


class CommandError(StandardError):

    def __init__(self, cmd, retcode, out, err):
        joined_cmd = ' '.join(cmd)
        msg = '''
  command: {joined_cmd}
  retcode: {retcode}
  output: {out}
  error: {err}
'''.format((), **locals())
        super(CommandError, self).__init__(msg)

def run_subprocess(args):
    proc = subprocess.Popen(args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            stdin=subprocess.PIPE)
    out, err = proc.communicate()
    if proc.returncode:
        raise CommandError(args, proc.returncode, out, err)
    return proc.returncode, out, err

def pipe_subprocess(args):
    proc = subprocess.Popen(args,
                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    return proc


class VM(object):

    def __init__(self, conf):
        self.conf = conf

    def run_local(self, args):
        proc = subprocess.Popen(args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                stdin=subprocess.PIPE)
        out, err = proc.communicate()
        if proc.returncode:
                raise CommandError(args, proc.returncode, out, err)
        return proc.returncode, out, err



class Kvm(VM):

    def __init__(self, conf):
        super(Kvm, self).__init__(conf)
        # Seed files
        self._meta_data_file = None
        self._user_data_file = None
        self._seed_path = None
        self._disk_image_path = None

    def _download_in_cache(self, source_url, name):
        """Download ``name`` from ``source_url`` in ``vm.download_cache``.

        :return: False if the file is in the download cache, True if a download
            occurred.
        """
        source = urlutils.join(source_url, name)
        download_dir = self.conf.get('vm.download_cache')
        target = os.path.join(download_dir, name)
        if not os.path.exists(target):
            run_subprocess(['wget', '--progress=dot:mega','-O', target, source])
            return True
        else:
            return False

    def download_iso(self):
        """Download the iso to install the vm.

        :return: False if the iso is in the download cache, True if a download
            occurred.
        """
        return self._download_in_cache(self.conf.get('vm.iso_url'),
                                       self.conf.get('vm.iso_name'))

    def download_cloud_image(self):
        """Download the cloud image to install the vm.

        :return: False if the image is in the download cache, True if a
            download occurred.
        """
        return self._download_in_cache(self.conf.get('vm.cloud_image_url'),
                                       self.conf.get('vm.cloud_image_name'))

    def _create_tempfile_with_content(self, content):
        temp = tempfile.NamedTemporaryFile()
        # We don't close the file here. This ensures it will be deleted when
        # self goes out of scope freeing us to handle the cleanup.
        temp.file.write(content)
        temp.file.flush()
        return temp

    def create_meta_data(self):
        self._meta_data_file = self._create_tempfile_with_content(
            self.conf.get('vm.meta_data'))

    def create_user_data(self):
        self._user_data_file = self._create_tempfile_with_content(
            self.conf.get('vm.user_data'))

    def create_seed(self):
        if self._meta_data_file is None:
            self.create_meta_data()
        if self._user_data_file is None:
            self.create_user_data()
        # FIXME: seed_path leak (no big deal) -- vila 2013-02-01
        seed_path = os.path.join(self.conf.get('vm.images_dir'),
                                 self.conf.expand_options('{vm.name}.seed'))
        run_subprocess(
            # We create the seed in the ``vm.images_dir`` directory, so
            # ``sudo`` is required
            ['sudo',
             'genisoimage', '-output', seed_path, '-volid', 'cidata',
             '-joliet', '-rock', '-input-charset', 'default',
             '-graft-points',
             'user-data=%s' % (self._user_data_file.name,),
             'meta-data=%s' % (self._meta_data_file.name,),
             ])
        self._seed_path = seed_path

    def create_image_from_cloud(self):
        cloud_image_path = os.path.join(self.conf.get('vm.download_cache'),
                                        self.conf.get('vm.cloud_image_name'))
        disk_image_path = os.path.join(
            self.conf.get('vm.images_dir'),
            self.conf.expand_options('{vm.name}.qcow2'))
        run_subprocess(
            ['sudo', 'qemu-img', 'convert',
             '-O', 'qcow2', cloud_image_path, disk_image_path])
        run_subprocess(
            ['sudo', 'qemu-img', 'resize',
             disk_image_path, self.conf.get('vm.disk_size')])
        self._disk_image_path = disk_image_path

    def run_pipe(self, args):
        proc = subprocess.Popen(args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        return proc

    def install_with_seed(self):
        if self._seed_path is None:
            self.create_seed()
        if self._disk_image_path is None:
            self.create_image_from_cloud()
        self.install_proc = self.run_pipe(
            ['sudo', 'virt-install',
             '--noautoconsole', # We'll connect later ourselves
             # Anticipate that we'll need a graphic card defined
             '--video', 'cirrus',
             '--name', self.conf.get('vm.name'),
             '--ram', self.conf.get('vm.ram_size'),
             '--vcpus', self.conf.get('vm.cpus'),
             '--disk', 'path=%s,format=qcow2' % (self._disk_image_path,),
             '--disk', 'path=%s' % (self._seed_path,),
             '--boot', 'hd', '--hvm',
             ])
        # Urgh, had to wait for virt-install to create the vm definition.
        time.sleep(2.0)
        # While `install` is running, let's connect to the console
        console_cmd = ['virsh', 'console', self.conf.get('vm.name')]
        console = self.run_pipe(console_cmd)
        lines = []
        while True:
            line = console.stdout.readline()
#            print "read: [%s]" % (line,) # so useful for debug...
            lines.append(line)
            if not line:
                # This should only happen when something goes wrong...
                raise CommandError(
                    console_cmd, console.returncode, '\n'.join(lines),
                    "EOF on console without expected final line")
            # FIXME: Use a regexp to be a bit less brittle (match on seed,
            # avoid the cloud-init version).

            # A typical install from seed on /dev/sdb looks like:

            # Cloud-init v. 0.7.1 finished at Tue, 05 Feb 2013 13:27:29 +0000. Datasource DataSourceNoCloudNet [seed=/dev/sdb][dsmode=net].  Up 7.86 seconds
            if line.startswith('Cloud-init v. 0.7.1 finished at'):
                break
        console.terminate()

    def poweroff(self):
        out, err = self.install_proc.communicate()
        return run_subprocess(
            ['sudo', 'virsh', 'destroy', self.conf.get('vm.name')])

    def undefine(self):
        return run_subprocess(
            ['sudo', 'virsh', 'undefine', self.conf.get('vm.name'),
             '--remove-all-storage'])

