from cStringIO import StringIO
import os
import sys

import testtools

import tests
from bin import setup_vm


class TestCaseWithHome(testtools.TestCase):

    def setUp(self):
        super(TestCaseWithHome, self).setUp()
        tests.set_cwd_to_tmp(self)
        tests.isolate_env(self)
        self.home_dir = os.path.join(self.test_base_dir, 'home')
        os.mkdir(self.home_dir)
        os.environ['HOME'] = self.home_dir


class TestVmMatcher(TestCaseWithHome):

    def setUp(self):
        super(TestVmMatcher, self).setUp()
        self.store = setup_vm.VmStore('.', 'foo.conf')
        self.matcher = setup_vm.VmMatcher(self.store, 'test')

    def test_empty_section_always_matches(self):
        self.store._load_from_string('foo=bar')
        matching = list(self.matcher.get_sections())
        self.assertEqual(1, len(matching))

    def test_specific_before_generic(self):
        self.store._load_from_string('foo=bar\n[test]\nfoo=baz')
        setup_vm.stopme = True
        matching = list(self.matcher.get_sections())
        setup_vm.stopme = False
        self.assertEqual(2, len(matching))
        # First matching section is for test
        self.assertEqual(self.store, matching[0][0])
        base_section = matching[0][1]
        self.assertEqual('test', base_section.id)
        # Second matching section is the no-name one
        self.assertEqual(self.store, matching[0][0])
        no_name_section = matching[1][1]
        self.assertIs(None, no_name_section.id)


class TestVmStack(TestCaseWithHome):

    def setUp(self):
        super(TestVmStack, self).setUp()
        self.conf = setup_vm.VmStack('foo')
        self.conf.store._load_from_string('''
vm.release=raring
vm.cpu_model=amd64
''')

    def test_raring_iso_url(self):
        self.assertEqual('http://cdimage.ubuntu.com/daily-live/current/',
                         self.conf.get('vm.iso_url'))

    def test_raring_iso_name(self):
        self.assertEqual('raring-desktop-amd64.iso',
                         self.conf.get('vm.iso_name'))

    def test_raring_cloud_image_url(self):
        self.assertEqual('http://cloud-images.ubuntu.com/raring/current/',
                         self.conf.get('vm.cloud_image_url'))

    def test_raring_cloud_image_name(self):
        self.assertEqual('raring-server-cloudimg-amd64-disk1.img',
                         self.conf.get('vm.cloud_image_name'))


class TestDownload(TestCaseWithHome):

    def setUp(self):
        # Downloading real isos or images is too long for tests, instead, we
        # fake it by downloading a small but known to exist file: MD5SUMS
        super(TestDownload, self).setUp()
        download_cache = os.path.join(self.test_base_dir, 'downloads')
        os.mkdir(download_cache)
        self.conf = setup_vm.VmStack('foo')
        self.conf.store._load_from_string('''
vm.iso_name=MD5SUMS
vm.cloud_image_name=MD5SUMS
vm.release=raring
vm.cpu_model=amd64
vm.download_cache=%s
''' % (download_cache,))

    def test_download_iso(self):
        vm = setup_vm.Kvm(self.conf)
        self.assertTrue(vm.download_iso())
        # Trying to download again will find the file in the cache
        self.assertFalse(vm.download_iso())

    def test_download_cloud_image(self):
        vm = setup_vm.Kvm(self.conf)
        self.assertTrue(vm.download_cloud_image())
        # Trying to download again will find the file in the cache
        self.assertFalse(vm.download_cloud_image())

    def test_download_fail(self):
        self.conf.set('vm.iso_name', 'I-dont-exist')
        vm = setup_vm.Kvm(self.conf)
        self.assertRaises(setup_vm.CommandError, vm.download_iso)


class TestSeed(TestCaseWithHome):

    def setUp(self):
        super(TestSeed, self).setUp()
        self.conf = setup_vm.VmStack('foo')
        self.vm = setup_vm.Kvm(self.conf)
        images_dir = os.path.join(self.test_base_dir, 'images')
        os.mkdir(images_dir)
        self.conf.store._load_from_string('''
vm.name=foo
vm.ssh_pub_key=bar
vm.release=raring
vm.images_dir=%s
''' % (images_dir,))

    def assertPathCleanedUp(self, tmp_path):
        self.assertTrue(os.path.exists(tmp_path))
        # The temp file is deleted when the vm object is
        del self.vm
        self.assertFalse(os.path.exists(tmp_path))

    def test_create_meta_data(self):
        self.vm.create_meta_data()
        self.assertPathCleanedUp(self.vm._meta_data_file.name)

    def test_create_user_data(self):
        self.vm.create_user_data()
        self.assertPathCleanedUp(self.vm._user_data_file.name)

    def test_create_seed(self):
        self.assertTrue(self.vm._seed_path is None)
        self.vm.create_seed()
        self.assertFalse(self.vm._seed_path is None)
        self.assertTrue(os.path.exists(self.vm._seed_path))


class TestImageFromCloud(TestCaseWithHome):

    def setUp(self):
        super(TestImageFromCloud, self).setUp()
        self.conf = setup_vm.VmStack('foo')
        self.vm = setup_vm.Kvm(self.conf)
        images_dir = os.path.join(self.test_base_dir, 'images')
        os.mkdir(images_dir)
        download_cache_dir = os.path.join(self.test_base_dir, 'download')
        os.mkdir(download_cache_dir)
        self.conf.store._load_from_string('''
vm.name=foo
vm.ssh_pub_key=bar
vm.release=raring
vm.images_dir=%s
vm.download_cache=%s
vm.cloud_image_name=fake.img
vm.disk_size=1M
''' % (images_dir, download_cache_dir))

    def test_create_image(self):
        cloud_image_path = os.path.join(self.conf.get('vm.download_cache'),
                                        self.conf.get('vm.cloud_image_name'))
        # We need a fake cloud image that can be converted
        setup_vm.run_subprocess(
            ['sudo', 'qemu-img', 'create',
             cloud_image_path, self.conf.get('vm.disk_size')])
        self.assertTrue(self.vm._disk_image_path is None)
        self.vm.create_image_from_cloud()
        self.assertFalse(self.vm._disk_image_path is None)
        self.assertTrue(os.path.exists(self.vm._disk_image_path))


class TestInstallWithSeed(TestCaseWithHome):

    def setUp(self):
        # We need a pre-seeded download cache from the user running the tests
        # as downloading the cloud image is too long.
        user_conf = setup_vm.VmStack(None)
        download_cache = user_conf.get('vm.download_cache')
        if download_cache is None:
            self.skip('vm.download_cache is not set')
        # We use some known reference
        reference_cloud_image_name = 'raring-server-cloudimg-amd64-disk1.img'
        cloud_image_path = os.path.join(
            download_cache, reference_cloud_image_name)
        if not os.path.exists(cloud_image_path):
            self.skip('%s is not available' % (cloud_image_path,))
        super(TestInstallWithSeed, self).setUp()
        # We need to allow other users to read this dir
        os.chmod(self.test_base_dir, 0755)
        # We also need to sudo rm it as root created some files there
        self.addCleanup(
            setup_vm.run_subprocess,
            ['sudo', 'rm', '-fr',
             os.path.join(self.test_base_dir, 'home', '.virtinst')])
        self.conf = setup_vm.VmStack('selftest')
        self.vm = setup_vm.Kvm(self.conf)
        images_dir = os.path.join(self.test_base_dir, 'images')
        os.mkdir(images_dir, 0755)
        # Setup user-data with a config that leads to a very short install time
        user_data = '''\
#cloud-config
locale: en_US.UTF-8
password: {vm.password}
chpasswd: { expire: False }
apt_update: false
apt_upgrade: false
'''
        self.conf.store._load_from_string('''
vm.name=selftest
vm.cpus=2,
vm.ssh_pub_key=bar
vm.release=raring
vm.images_dir=%s
vm.download_cache=%s
vm.cloud_image_name=%s
vm.disk_size=8G
vm.user_data="""%s"""
''' % (images_dir, download_cache, reference_cloud_image_name, user_data))

    def get_vm_states(self):
        retcode, out, err = setup_vm.run_subprocess(['virsh', 'list', '--all'])
        lines = out.splitlines()
        # Get rid of header/footer
        lines = lines[2:-1]
        vm_states = {}
        for line in lines:
            caret, name, state = line.split(None, 2)
            vm_states[name] = state
        return vm_states

    def assertVmState(self, expected):
        vm_states = self.get_vm_states()
        self.assertEqual(expected, vm_states[self.vm.conf.get('vm.name')])

    def test_install_from_seed(self):
        self.addCleanup(self.vm.undefine)
        self.vm.install_with_seed()
        self.assertVmState('running')
        self.vm.poweroff()
        self.assertVmState('shut off')
